use crate::token::Token;
use logos::{Lexer, Logos};
use peekmore::{PeekMore, PeekMoreIterator};
use std::fmt;
use std::str::Chars;

pub struct Scanner<'a> {
    lexer: Lexer<'a, Token>,
    line: u16,
}

impl<'a> fmt::Display for Scanner<'a> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{} '{}'", self.line, self.lexer.slice())
    }
}

impl<'a> Scanner<'a> {
    pub fn new(source: &'a str) -> Self {
        Scanner {
            line: 1,
            lexer: Token::lexer(source),
        }
    }
    pub fn next(&mut self) -> Option<Result<Token, ()>> {
        self.lexer.next()
    }

    pub fn compile(&mut self) {
        let mut line = 1;
        while let Some(Ok(token)) = self.lexer.next() {
            match token {
                Token::TokenNewline => {
                    line += 1;
                    print!("{:4}", line);
                }
                _ => {
                    if self.lexer.span().next().unwrap() == 0 {
                        print!("{:4}", line);
                    } else {
                        print!("   |");
                    }
                }
            }
            println!(
                ":{:<3} {:20} {:20}",
                self.lexer.span().next().unwrap(),
                token,
                format!("{:?}", self.lexer.slice())
            );
        }
    }
}

#[cfg(test)]
mod tests {

    use crate::token::Token;
    use logos::Logos;
    #[test]
    fn single_chars() {
        //assert_token(String::from(""), Token::TokenEof);
        assert_token(String::from("("), Token::TokenLeftParen);
        assert_token(String::from("}"), Token::TokenRightBrace);
        assert_token(String::from("-"), Token::TokenMinus);
        assert_token(String::from("+"), Token::TokenPlus);
        assert_token(String::from("/"), Token::TokenSlash);
    }

    #[test]
    fn double_chars() {
        assert_token(String::from("=="), Token::TokenEqualEqual);
        assert_token(String::from("!="), Token::TokenBangEqual);
        assert_token(String::from(">"), Token::TokenGreater);
        assert_token(String::from(">="), Token::TokenGreaterEqual);
    }
    #[test]
    fn strings() {
        assert_token_lexeme(String::from("\"mox\""), Token::TokenString, "\"mox\"");
        assert_token_lexeme(String::from("\"\""), Token::TokenString, "\"\"");
    }

    #[test]
    fn numbers() {
        assert_token_lexeme(String::from("0"), Token::TokenNumber(0.0), "0");
        assert_token_lexeme(String::from("4"), Token::TokenNumber(4.0), "4");
        assert_token_lexeme(String::from("42"), Token::TokenNumber(42.0), "42");
        assert_token_lexeme(String::from("13.99"), Token::TokenNumber(13.99), "13.99");
    }
    #[test]
    fn newlines() {
        assert_tokens(
            String::from("+\n//comment\n-"),
            &vec![
                Token::TokenPlus,
                Token::TokenNewline,
                Token::TokenComment,
                Token::TokenNewline,
                Token::TokenMinus,
            ],
        );
    }

    #[test]
    fn identifier() {
        assert_token(String::from("class"), Token::TokenClass);
        assert_token(String::from("if"), Token::TokenIf);
        assert_token(String::from("while"), Token::TokenWhile);
        assert_token(String::from("true"), Token::TokenTrue);
        assert_token(String::from("false"), Token::TokenFalse);

        assert_token(String::from("cap"), Token::TokenFalse);

        assert_token_lexeme(String::from("mox"), Token::TokenIdentifier, "mox");
    }
    fn assert_token(source: String, expected: Token) {
        let mut lex = Token::lexer(&source);
        let token = lex.next();

        assert_eq!(token, Some(Ok(expected)));
    }

    fn assert_token_lexeme(source: String, expected_type: Token, expected_lexeme: &str) {
        let mut lex = Token::lexer(&source);
        let token = lex.next();

        assert_eq!(token, Some(Ok(expected_type)));

        assert_eq!(lex.slice(), expected_lexeme);
    }
    fn assert_tokens(source: String, expected_tokens: &Vec<Token>) {
        let mut lex = Token::lexer(&source);
        for expected in expected_tokens {
            let actual = lex.next();
            assert_eq!(actual, Some(Ok(*expected)));
        }
    }
}
